/**
 * @file fedkit-private.h
 * Private structures for fedkit implementation
 */
/* fedkit documentatio is at http://cern.ch/cano/fedkit/ */
/* to compile the program, from directory itools, execute the following command : */
/*
 Maintainer : Eric Cano, CERN-EP/CMD <Eric.Cano@cern.ch>
 
 $Id: fedkit-private.h,v 1.1 2007/10/09 08:19:01 cano Exp $
*/
#ifndef _FEDKIT_PRIVATE_H_
#define _FEDKIT_PRIVATE_H_

#include "../include/fedkit-types.h"

/* PCI identification of board */
#define _FK_VENDOR              (0xECD6)
#define _FK_DEVICE              (0xFD05)

/* offsets */
#define _FK_BKSZ_OFFSET         (0x80)
#define _FK_WCFADDR_OFFSET      (0x84)
#define _FK_WCFR_OFFSET         (0x08)
#define _FK_WCFWADDR_OFFSET     (0x88)
#define _FK_FRBKFIFO_OFFSET     (0x04)
#define _FK_CSR_OFFSET          (0x00)
#define _FK_DEBUG_WRITE0        (0x40)
#define _FK_FIFOSTAT_OFFSET     (0x110)
#define _FK_DUMPDATA0_OFFSET    (0x100)
#define _FK_DUMPDATA1_OFFSET    (0x104)
#define _FK_DUMPCTRL0_OFFSET    (0x108)
#define _FK_DUMPCTRL1_OFFSET    (0x10C)


#define _FK_BAR0_range          (0x8C)
#define _FK_BAR1_range          (0x100)

/* CSR bits*/
#define _FK_CSR_HALF_EMPTY_INT_STATUS           (0x00000002)
#define _FK_CSR_EMPTY_INT_STATUS                (0x00000001)
#define _FK_CSR_HALF_EMPTY_INT_ENABLE           (0x00000200)
#define _FK_CSR_EMPTY_INT_ENABLE                (0x00000100)
#define _FK_CSR_ALL_INTS_ENABLE                 (0x00000300)
#define _FK_CSR_SLINK_DOWN                      (0x00004000)
#define _FK_CSR_SLINK_UP                        (0x00008000)
#define _FK_CSR_ABORT                           (0x00010000)
#define _FK_CSR_SOFT_RESET                      (0x00020000)
#define _FK_CSR_LINK_CTRL                       (0x00040000)
#define _FK_CSR_AUTOSEND_FULL                   (0x00080000)
#define _FK_CSR_LINK0_ENABLE                    (0x01000000)
#define _FK_CSR_LINK1_ENABLE                    (0x02000000)
#define _FK_CSR_ALL_LINKS_ENABLE                (0x03000000)
#define _FK_CSR_DEFAULT_LINK_PATTERN            _FK_CSR_LINK0_ENABLE
#define _FK_CSR_LINK_IS_UP                      (0x80000000)
#define _FK_CSR_LINK_DEBUG                      (0x10000000)

#define _FK_CSR_LINKS_SHIFT                     (24)

/* How much time has to be slept in between link down and link up */
#define _FK_SLINK_SLEEP                         (1)

/* FIFO_STAT bits */
#define _FK_FIFOSTAT_LINK0                        (1 << 0)
#define _FK_FIFOSTAT_LINK1                        (1 << 1)
#define _FK_FIFOSTAT_PCI                          (1 << 2)

/* DUMPCTRL0 bits */
#define _FK_DUMPCTRL0_K                           (1 << 0)

/* FIFO sizes */
#define _FK_free_data_fifo_depth        (1024)
#define _FK_wc_fifo_depth               (1024)
#define _FK_wc_fifo_address_mask        (0x000003FF)
#define _FK_wc_fifo_mask                (0x000007FF)
#define _FK_wc_fifo_extension_mask      (0x00000400)

/* Block sizes */
#define FK_MAX_BLOCK_SIZE               (64*1024)
#define FK_MIN_BLOCK_SIZE               (512)
#define FK_SIZE_ALIGN_MASK              (0xFFFFFFF8)

/* default values */
#define FK_DEFAULT_BLOCK_SIZE           (4096)
#define FK_DEFAULT_BLOCK_NUMBER         (2048)
#define FK_DELAULT_HEADER_SIZE          (0)

/* sender related */
#define _FK_SENDER_VENDOR               (0xECD6)
#define _FK_SENDER_DEVICE               (0xFE01)

#define _FK_sender_BAR0_range           (0x0000C)
#define _FK_sender_BAR1_range           (0x10000)
#define _FK_sender_BAR2_range           (0x00008)

#define _FK_SENDER_MEVTADDR             (0x0)
#define _FK_SENDER_MEVTSZ               (0x4)
#define _FK_SENDER_MCTRL                (0x8)

#define _FK_SENDER_SCTRL                (0x0)
#define _FK_SENDER_SDATA                (0x4)

#define _FK_SENDER_GCTRL                (0x0)
#define _FK_SENDER_GEVENT               (0x4)

#define _FK_SENDER_FIFO_EMPTY           (0x00000001)
#define _FK_SENDER_FIFO_HALF_FULL       (0x00000002)
#define _FK_SENDER_FIFO_FULL            (0x00000004)
#define _FK_SENDER_MASTER_DMA_DONE      (0x00000008)

#define _FK_SENDER_ENABLE_FIFO_EMPTY            (0x00000100)
#define _FK_SENDER_ENABLE_FIFO_HALF_FULL        (0x00000200)
#define _FK_SENDER_ENABLE_FIFO_FULL             (0x00000400)
#define _FK_SENDER_ENABLE_FIFO_DMA_DONE         (0x00000800)

#define _FK_SENDER_GENERATOR_MODE       (0x00010000)
#define _FK_SENDER_DMA_INT_CLEAR        (0x00020000)
#define _FK_SENDER_DMA_IN_PROGRESS      (0x00040000)
#define _FK_SENDER_SOFT_RESET           (0x00080000)

#define _FK_SENDER_FEDID_SHIFT    		(16)

/* HEADER positions (in 32 bits words..) */
#define _FKH_SHIFT_Evt_type     (24)
#define _FKH_WORD_Evt_type              (1)
#define _FKH_MASK_Evt_type              (0xF)

#define _FKH_SHIFT_LV1_id               (0)
#define _FKH_WORD_LV1_id                (1)
#define _FKH_MASK_LV1_id                (0xFFFFFF)

#define _FKH_SHIFT_BX_id                (20)
#define _FKH_WORD_BX_id                 (0)
#define _FKH_MASK_BX_id                 (0xFFF)

#define _FKH_SHIFT_Source_id    (8)
#define _FKH_WORD_Source_id             (0)
#define _FKH_MASK_Source_id             (0xFFF)

#define _FKH_SHIFT_FOV                  (4)
#define _FKH_WORD_FOV                   (0)
#define _FKH_MASK_FOV                   (0xF)

/* TRAILER positions (in 32 bits words..) */
#define _FKT_SHIFT_Evt_lgth             (0)
#define _FKT_WORD_Evt_lgth              (1)
#define _FKT_MASK_Evt_lgth              (0xFFFFFF)

#define _FKT_SHIFT_CRC                  (16)
#define _FKT_WORD_CRC                   (0)
#define _FKT_MASK_CRC                   (0xFFFF)

#define _FKT_SHIFT_Evt_stat             (8)
#define _FKT_WORD_Evt_stat              (0)
#define _FKT_MASK_Evt_stat              (0xF)

#define _FKT_SHIFT_TTS                  (4)
#define _FKT_WORD_TTS                   (0)
#define _FKT_MASK_TTS                   (0xF)

struct _FK_data_fifo {
    int read;
    int write; 
    int size;
    int done_read;
    int done_write;
    int int_disabled;
};

/* Some status is encoded by the FEDkit in the receied element.*/
struct _FK_wc_fifo_element {
    uint32_t    wc:24;
    uint8_t     :4;
    uint8_t     no_fragment:1;
    uint8_t	FEDCRC_error:1;
    uint8_t     truncated:1;
    uint8_t     CRC_error:1;
};

struct _FK_wc_fifo {
    struct _FK_wc_fifo_element fifo_elements[_FK_wc_fifo_depth];
    uint32_t read;
    uint32_t write;
};

struct _FK_free_buffer_FIFO
{
};


#ifndef __KERNEL__
#include <sys/ioctl.h>
#endif

/* First (tentative) implementation of a native-linux-x86 version of the 
   fedkit. Allows things to be cleaner on a Linux system, at the cost of portability
 */

/*
 * Driver operations work the following way :
 *  - read and write not implemented
 *  - open only allowed once per device
 *  - minor number identifies senders/receivers. 32 receivers maximum (should be more than enough).
 *        minor above 32 means sender.
 *  - one ioctl read all base addresses
 *  - a mmap operation for those base addresses makes the fedkit_open or fedkit_sender_open operation ready
 *  - a setget receiver parameters ioctl allows driver/board and userland to agree on the parameters
 *  - a get receiver memory blocks, so that the receiver get the bigphys zones allocated by the kernel
 *  - a second mmap operation to mmap the bigphys zones into the process'es space.
 *  - still have to determine how to make both mmap opearations work with the same file descriptor...
 */


#define MAX_FEDKIT_SENDER_NUMBER (32)

struct _fedkit_board_addresses {
    uint32_t base_address_0;
    uint32_t base_address_1;
    uint32_t base_address_2;
    uint32_t FPGA_version;
};

struct _fedkit_receiver_parameters {
    uint32_t block_number;
    uint32_t block_size;
    uint32_t do_allocate_blocks;
};

struct _fedkit_memory_blocks {
    void * free_blocks_fifo_kernel;
    void * free_blocks_fifo_physical;
    void * data_blocks_kernel;
    void * data_blocks_physical;
    void * wc_fifo_kernel;
    void * wc_fifo_physical;
};

#ifndef FEDKIT_NO_MASTER_INTS
#define _FK_MASTER_FIFO_SIZE (10)
struct _fedkit_sender_master_data_t {
    void * address;
    int word_count;
};

struct _fedkit_sender_master_fifo_t {
    int write;
    int read;
    int size;
    int int_disabled;
    int first_dma_done;
    void * inprogress;
    struct _fedkit_sender_master_data_t data[_FK_MASTER_FIFO_SIZE];
};

struct _fedkit_sender_memory_blocks {
    void * master_fifo_kernel;
    void * master_fifo_physical;
};

/* Duplication from fedkit.h (with name change (_) ) */
struct _fedkit_dbuff {
    uint32_t size;
    uint64_t physical_address;
    void * kernel_address;
    void * user_address;
};

#ifdef __KERNEL__ /* type definitions common to the driver files */

#define FEDKIT_DEVNAME_SIZE (20)

/* list of all the allocated receivers */
struct kfedkit_receiver_t
{
    int index;
    char device_name [FEDKIT_DEVNAME_SIZE];
    uint32_t base_address0;    
    uint32_t base_address1;    
    uint32_t base_address2;
    uint32_t * map0;
    uint32_t * map1;
    uint32_t * map2;
    uint32_t FPGA_version;
    int block_number;
    int block_size;
    int do_allocate_blocks;
    int interrupt_registered;
    void * block_fifo;
    void * blocks;
    void * wc_fifo;
    struct pci_dev * device;
    struct semaphore lock;
    struct kfedkit_receiver_t *next;
    struct _FK_data_fifo * data_fifo;
    void **  data_fifo_tab;
};

struct kfedkit_sender_t 
{
    int index;
    char device_name [FEDKIT_DEVNAME_SIZE];
    uint32_t base_address0;    
    uint32_t base_address1;    
    uint32_t base_address2;
    uint32_t * map0;
    uint32_t * map1;
    uint32_t * map2;
    uint32_t FPGA_version;
    struct pci_dev * device;
    struct semaphore lock;
#ifndef FEDKIT_NO_MASTER_INTS
    struct _fedkit_sender_master_fifo_t *master_fifo;
    int interrupt_registered;
#endif /* ndef FEDKIT_NO_MASTER_INTS */
    struct kfedkit_sender_t *next;
};

enum kfedkit_type {
    sender_board,
    receiver_board
};

struct kfedkit_board_t {
    struct kfedkit_dbuff_t  * dbuffs;
    enum kfedkit_type type;
    union {
        struct kfedkit_sender_t sender;
        struct kfedkit_receiver_t receiver;
    };
};

/* structure used to keep track of the dbuffs allocated to a file descriptor */
struct kfedkit_dbuff_t {
	void * kernel_address;
	struct kfedkit_dbuff_t * next;
};


int kfedkit_ioctl_alloc_dbuff (struct inode *inode_p, struct file *f,
			       unsigned int cmd, unsigned long arg);

int kfedkit_ioctl_free_dbuff (struct inode *inode_p, struct file *f,
			      unsigned int cmd, unsigned long arg);

void kfedkit_dbuffs_free (struct kfedkit_dbuff_t *dbuffs);

int kfedkit_Dbuff_mmap (struct file *file_p, struct vm_area_struct *vma);



#endif /* def _KERNEL_ */

#endif /* ndef FEDKIT_NO_MASTER_INTS */

#define FEDKIT_MAJOR                (125)
#define FEDKIT_IOCTL_MAGIC          ('Y')

#define FEDKIT_GET_BOARD_ADDRESSES       _IOR  (FEDKIT_IOCTL_MAGIC, 0, struct _fedkit_board_addresses)
#define FEDKIT_SETGET_RECEIVER_PARAMS    _IOWR (FEDKIT_IOCTL_MAGIC, 0, struct _fedkit_receiver_parameters)
#define FEDKIT_GET_RECEIVER_MEMBLOCKS    _IOR  (FEDKIT_IOCTL_MAGIC, 1, struct _fedkit_memory_blocks)
#define FEDKIT_ALLOC_DBUFF               _IOWR (FEDKIT_IOCTL_MAGIC, 1, struct _fedkit_dbuff)
#define FEDKIT_FREE_DBUFF                _IOWR (FEDKIT_IOCTL_MAGIC, 2, struct _fedkit_dbuff)

#ifndef FEDKIT_NO_MASTER_INTS
#define FEDKIT_GET_SENDER_MEMBLOCKS      _IOR  (FEDKIT_IOCTL_MAGIC, 1, struct _fedkit_sender_memory_blocks)
#endif /* ndef FEDKIT_NO_MASTER_INTS */

#endif /* ifndef _FEDKIT_PRIVATE_H_ */







